#include "tool.h"
#include <curses.h>
#include <strings.h>

//==========================
//        PRIVATE
//==========================
int possibility_count = 0;
char **remaining_bank;
int remaining_bank_count;
possibility_t **possibilities;

/**
 * @brief Destroiy a given matches set
 *
 * @param matches
 */
void destroy_matches(matches_t *matches) {
    for (int i = 0; i < matches->count; i++)
        free(matches->words[i]);
    free(matches->words);
    free(matches);
}

/**
 * @brief Filter the remaining bank
 *
 * @param word
 * @param pattern
 * @return
 */
matches_t *filter_out_remaining_bank(char *word, Pattern *pattern) {
    matches_t *matches = malloc(sizeof(matches_t));
    matches->count = 0;

    // Create empty word matched array
    char tmp[remaining_bank_count][WORD_LENGHT + 1];
    for (int i = 0; i < remaining_bank_count; i++)
        strcpy(tmp[i], "\0");

    // Check validity for each remaning word
    for (int j = 0; j < remaining_bank_count; j++) {
        char current_word[WORD_LENGHT + 1];
        strcpy(current_word, remaining_bank[j]);

        // Exclude current word
        if (strcmp(current_word, word) == 0)
            continue;

        bool valid_word = true;
        for (int i = 0; i < WORD_LENGHT; i++) {
            // Filter out the non-valid words
            if ((pattern[i] == CORRECT && (int)(index(current_word, word[i]) - current_word) != i) ||
                (pattern[i] == MISPLACED && strchr(current_word, word[i]) == NULL) ||
                (pattern[i] == WRONG && strchr(current_word, word[i]) != NULL)) {
                valid_word = false;
                break;
            }

            // Clear the current char to not impact futur letter checks
            if (pattern[i] == CORRECT)
                current_word[i] = ' ';
        }

        if (!valid_word)
            continue;

        strcpy(tmp[j], remaining_bank[j]);
        matches->count++;
    }

    matches->words = malloc(matches->count * sizeof(char *));
    int idx = 0;
    for (int i = 0; i < remaining_bank_count; i++) {
        if (strlen(tmp[i]) != 0) {
            //            printf("%s\n", tmp[i]);
            matches->words[idx] = calloc(WORD_LENGHT + 1, sizeof(char));
            strcpy(matches->words[idx], tmp[i]);
            idx++;
        }
    }

    return matches;
}

/**
 * @brief Compute the entropy of a given word
 *
 * @param word
 * @return
 */
double compute_entropy(char *word) {
    double sum = 0.0;

    Pattern p[4] = {WRONG, MISPLACED, CORRECT, NAP};
    Pattern *list[WORD_LENGHT] = {p, p, p, p, p};
    Pattern *curr[WORD_LENGHT];

    int n = sizeof(list) / sizeof(*list);
    bool done = false;

    for (int i = 0; i < n; i++)
        curr[i] = list[i];

    // Compute the product itself (0, 0, 0, 0, 0), (1, 0, 0, 0, 0), (2, 0, 0, 0, 0), (0, 1, 0, 0, 0), etc...
    // O(Patterns^WORD_LENGTH)
    while (!done) {
        // Compute current pattern entropy
        matches_t *matches = filter_out_remaining_bank(word, (Pattern[WORD_LENGHT]){*curr[0], *curr[1], *curr[2], *curr[3], *curr[4]});

        // Skip unmatch pattern
        if (matches->count != 0) {
            // Apply Shannon's formula
            double px = ((double)matches->count / (double)remaining_bank_count);
            sum += px * (log2(1 / px));
        }

        destroy_matches(matches);

        // Go to next pattern
        int pattern_inc = 0;
        curr[pattern_inc]++;

        // Loop over the pattern and increment next position
        while (*curr[pattern_inc] == NAP) {
            curr[pattern_inc] = list[pattern_inc];
            pattern_inc++;

            if (pattern_inc == n) {
                done = true;
                break;
            }

            curr[pattern_inc]++;
        }
    }

    return sum;
}

/**
 * @brief Create a possibility
 *
 * @param word
 * @param entropy
 * @return
 */
possibility_t *create_possibility(char *word) {
    possibility_t *p = malloc(sizeof(possibility_t));
    p->word = calloc(WORD_LENGHT + 1, sizeof(char));
    strcpy(p->word, word);
    p->entropy = compute_entropy(word);
    return p;
}

/**
 * @brief Compare two given possibilites
 *
 * @param p1
 * @param p2
 * @return
 */
int compare_possibilities(const void *s1, const void *s2) {
    const possibility_t *p1 = *(possibility_t **)s1;
    const possibility_t *p2 = *(possibility_t **)s2;

    if (p1->entropy < p2->entropy)
        return -1;
    else if (p1->entropy > p2->entropy)
        return 1;
    else
        return 0;
}

//==========================
//        PUBLIC
//==========================
void init_tool() {
    remaining_bank_count = get_bank_size();
    remaining_bank = malloc(remaining_bank_count * sizeof(char *));
    possibility_count = remaining_bank_count;
    possibilities = NULL;

    char **tmp = get_bank();
    for (int i = 0; i < remaining_bank_count; i++) {
        remaining_bank[i] = malloc((WORD_LENGHT + 1) * sizeof(char));
        strcpy(remaining_bank[i], tmp[i]);
    }
}

void destroy_tool() {
    for (int i = 0; i < remaining_bank_count; i++)
        free(remaining_bank[i]);

    free(remaining_bank);

    if (possibilities != NULL)
        destroy_possibilities(possibilities);
}

void destroy_possibilities(possibility_t **p) {
    for (int i = 0; i < possibility_count; i++) {
        free(p[i]->word);
        free(p[i]);
    }
    free(p);
}

bool string_contains(char *foo, char c) {
    char sc[2] = {c, '\0'};
    return (int)strcspn(foo, sc) != WORD_LENGHT;
}

possibility_t **compute_matches(char *word, Pattern *pattern) {
    matches_t *matches = filter_out_remaining_bank(word, pattern);

    if (matches->count == 0) {
        destroy_matches(matches);
        return NULL;
    }

    // Create the new possibilities list
    // Delete no longer used words
    if (possibility_count != get_bank_size()) {
        for (int i = 0; i < possibility_count; i++) {
            free(possibilities[i]->word);
            free(possibilities[i]);
        }
    }

    possibility_count = matches->count;
    possibilities = realloc(possibilities, possibility_count * sizeof(possibility_t *));

    for (int i = 0; i < possibility_count; i++)
        possibilities[i] = create_possibility(matches->words[i]);

    save_computed_matches(possibilities);
    destroy_matches(matches);

    // Sort possibilities by entropy ASC
    qsort(possibilities, possibility_count, sizeof(possibility_t *), compare_possibilities);

    return possibilities;
}

void save_computed_matches(possibility_t **matches) {
    // Free non-matching words
    for (int i = possibility_count; i < remaining_bank_count; i++)
        free(remaining_bank[i]);

    // Realloc for the new matching words
    remaining_bank_count = possibility_count;
    remaining_bank = realloc(remaining_bank, remaining_bank_count * sizeof(char *));
    
    for (int i = 0; i < remaining_bank_count; i++)
        strcpy(remaining_bank[i], matches[i]->word);
}

int get_remaining_bank_count() { return remaining_bank_count; }

char **get_remaining_bank() { return remaining_bank; }
