#include "wordle.h"
#include <curses.h>
#include <stdlib.h>

//==========================
//        PRIVATE
//==========================
Gamemode mode;
bool _game_finished = false;
int score = 0;
int *scores;
bool p2_turn = false;
int turn = 0;
char **tries;
char *chosen_word;
Pattern **patterns;
possibility_t **remaining_possibilities;
int current_try_id = 0;
int current_try_letter_id = 0;

/**
 * @brief Render the game for the selected gamemode
 */
void render_game() {
    switch (mode) {
    case SOLO:
        draw_title(SOLO_TITLE_ID);
        break;

    case VERSUS:
        draw_title(ONE_V_ONE_TITLE_ID);
        break;

    case TOOL_ASSISTED:
        draw_title(COMPUTER_TITLE_ID);
        break;
    }
    
    show_gameboard(TRIES_COUNT, WORD_LENGHT, tries, patterns, scores, mode, p2_turn, turn);

    move(LINES - 2, COLS - 10);
}

/**
 * @brief Handle controls
 *
 * @param key
 */
void handle_controls(int key) {
    switch (key) {
    case KEYBIND_HELP:
        toggle_help();
        break;

    case KEYBIND_QUIT:
        terminate_game();
        break;

    case KEYBIND_GUESS:
        if (!validate_guess(tries[current_try_id]))
            break;

        // Patterns
        Pattern *pattern = generate_pattern();
        for (int i = 0; i < WORD_LENGHT; i++)
            patterns[current_try_id][i] = pattern[i];
        free(pattern);
        pattern = NULL;

        // Place correct letters from previous in the current one
        if (current_try_id < TRIES_COUNT && strcmp(tries[current_try_id], get_answer()) != 0) {
            for (int i = 0; i < WORD_LENGHT; i++) {
                if (patterns[current_try_id][i] == CORRECT)
                    tries[current_try_id][i] = tries[current_try_id][i];
            }
        } 

        if (get_remaining_bank_count() > 0 && mode == TOOL_ASSISTED)
            remaining_possibilities = compute_matches(tries[current_try_id], patterns[current_try_id]);

        current_try_id += 1;
        current_try_letter_id = 0;

        if (mode == TOOL_ASSISTED && remaining_possibilities != NULL) {
            show_tool();

            if (current_try_id == 0 || get_remaining_bank_count() <= 0)
                break;

            char words[get_remaining_bank_count()][WORD_LENGHT + 1];
            double entropies[get_remaining_bank_count()];

            for (int i = 0; i < get_remaining_bank_count(); i++) {
                strcpy(words[i], remaining_possibilities[i]->word);
                entropies[i] = remaining_possibilities[i]->entropy;
            }

            update_tool(get_remaining_bank_count(), words, entropies);
        }
        break;
    }
}

//==========================
//        PUBLIC
//==========================
void initialize_game() {
    if (mode == TOOL_ASSISTED)
        init_tool();

    // Reset game variables
    _game_finished = false;
    current_try_id = 0;
    current_try_letter_id = 0;
    score = 0;
    scores = calloc(2, sizeof(int));
    chosen_word = malloc((WORD_LENGHT + 1) * sizeof(char));

    // Tries and answer setup
    set_answer(get_random_word());

    tries = calloc(TRIES_COUNT, sizeof(char *));
    patterns = malloc(sizeof(Pattern *) * TRIES_COUNT);

    for (int i = 0; i < TRIES_COUNT; i++) {
        patterns[i] = calloc(WORD_LENGHT + 1, sizeof(Pattern));
        tries[i] = calloc(WORD_LENGHT + 1, sizeof(char)); // +1 because of \0
    }
}

void terminate_game() { _game_finished = true; }

void destroy_game() {
    if (mode == TOOL_ASSISTED)
        destroy_tool();

    free(scores);
    free(chosen_word);

    for (int i = 0; i < TRIES_COUNT; i++) {
        free(tries[i]);
        free(patterns[i]);
    }

    free(tries);
    free(patterns);
}

void restart_game() {
    if (mode == TOOL_ASSISTED) {
        destroy_tool();
        init_tool();
    }

    // Versus mode scores or solo score
    if (mode == VERSUS) {
        if (turn >= VERSUS_TURN_COUNT) {
            show_winner(p2_turn, scores);
            getch();
            terminate_game();
            return;
        }

        scores[p2_turn] += (TRIES_COUNT - current_try_id) * 10;
        p2_turn = !p2_turn;

        if (!p2_turn)
            turn++;
    } else 
        scores[0] += (TRIES_COUNT - current_try_id) * 10;

    set_answer(get_random_word());

    current_try_id = 0;
    current_try_letter_id = 0;

    for (int i = 0; i < TRIES_COUNT; i++) {
        for (int j = 0; j < WORD_LENGHT; j++) {
            patterns[i][j] = WRONG;
            tries[i][j] = '\0';
        }
    }
}

bool lose_condition() { return (current_try_id > TRIES_COUNT - 1); }

bool win_condition() {
    for (int i = 0; i < WORD_LENGHT; i++) {
        if (patterns[current_try_id - (current_try_id > 0 ? 1 : 0)][i] != CORRECT)
            return false;
    }

    return true;
}

bool validate_letter(int key) { return (islower(key) || isupper(key)); }

Pattern *generate_pattern() {
    Pattern *pattern = calloc(WORD_LENGHT, sizeof(int));
    char cpy[WORD_LENGHT + 1];
    strcpy(cpy, chosen_word);

    for (int i = 0; i < WORD_LENGHT; i++) {
        if (strchr(cpy, tries[current_try_id][i]) == NULL)
            continue;

        if (tries[current_try_id][i] == chosen_word[i]) {
            pattern[i] = CORRECT;
            cpy[i] = ' ';
            continue;
        }

        char *addr;
        if ((addr = strchr(chosen_word, tries[current_try_id][i])) != NULL) {
            pattern[i] = MISPLACED;
            cpy[(int)(addr - chosen_word)] = ' '; // Get the index of the found char and replace with space
            continue;
        }
    }

    return pattern;
}

bool validate_guess(char current_try[WORD_LENGHT]) {
    // Length check
    if (strlen(current_try) < WORD_LENGHT)
        return false;

    // Presence check
    char **bank = get_bank();
    bool present = false;
    for (int i = 0; i < get_bank_size(); i++) {
        if (!strcmp(bank[i], current_try))
            present = true;
    }

    if (!present) {
        current_try_letter_id = 0;
        for (int i = 0; i < WORD_LENGHT; i++)
            tries[current_try_id][i] = '\0';
        return false;
    }

    return true;
}

void set_gamemode(Gamemode menu_gamemode) { mode = menu_gamemode; }

void set_answer(char answer[WORD_LENGHT]) { strcpy(chosen_word, answer); }

void set_try(char current_try[WORD_LENGHT]) { strcpy(tries[current_try_id], current_try); }

Gamemode get_gamemode() { return mode; }

char *get_answer() { return chosen_word; }

char *get_try() { return tries[current_try_id]; }

void launch_game() {
    initialize_game();

    while (!game_finished()) {
        // Render
        render_game();

        // Guard clause
        if (win_condition() || lose_condition()) {
            show_answer(chosen_word);
            restart_game();
            handle_controls(getch());
            hide_answer();
            continue;
        }

        // Key handling
        // || Need to check letter by letter every time to be able to continue using
        // || shortcut keys like Ctrl+h during a guess.
        int key = getch();

        if (validate_letter(key)) {
            if (current_try_letter_id < WORD_LENGHT) {
                tries[current_try_id][current_try_letter_id] = toupper(key);
                current_try_letter_id += 1;
            }
        }

        if (key == KEY_BACKSPACE) {
            if (current_try_letter_id > 0) {
                current_try_letter_id -= 1;
                tries[current_try_id][current_try_letter_id] = '\0';
            }
        }

        handle_controls(key);
        refresh();
    }

    destroy_game();
}

bool game_finished() { return _game_finished; }
